use crate::{
    errors::CvfsResult,
    secure_file::{ActiveFile, SecureFileHandler},
};
use multi_key_manager::{file::File, OpeningNonce};
use serde::{Deserialize, Serialize};
use std::{
    collections::HashMap,
    io::{Read, Result as IoResult},
    path::PathBuf,
};

pub type PathString = String;

#[derive(Serialize, Deserialize)]
pub struct Metadata {
    pub(crate) files: HashMap<PathString, FileInfo>,
}
impl Metadata {
    /// creates a new empty index
    pub fn empty() -> Self {
        Self {
            files: HashMap::new(),
        }
    }
    pub fn restore_from_slice(input: &[u8]) -> IoResult<Self> {
        todo!("Metadata::restore_from_slice")
    }

    pub fn find_info(&mut self, id: &FileTag) -> Option<&mut FileInfo> {
        self.files
            .values_mut()
            .filter(|element| match &element.active {
                Some(active) => active.id == *id,
                None => false,
            })
            .next()
    }
}

/// Information stored per file, except the path in the virtual file sytem
#[derive(Serialize, Deserialize)]
pub struct FileInfo {
    /// this exists if the file is already open
    #[serde(skip)]
    pub(crate) active: Option<ActiveFile>,
    pub(crate) real_path: PathBuf,
    /// keep this as private as possible
    nonce: u128,
}
impl FileInfo {
    pub fn new(real_path: PathBuf) -> Self {
        // TODO validate the path?
        Self {
            active: None,
            real_path,
            nonce: 1,
        }
    }

    pub fn get_opening_nonce(&self) -> OpeningNonce {
        OpeningNonce::new(self.nonce.clone())
    }
}
impl core::fmt::Debug for FileInfo {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("FileInfo")
            .field("active", &self.active.is_some())
            .field(
                "real_path",
                &format_args!(
                    "/blocks/{:?}",
                    self.real_path.file_name().unwrap_or_default()
                ),
            )
            .finish()
    }
}

#[derive(Clone, PartialEq, Eq)]
pub struct FileTag(pub(crate) usize);

#[cfg(test)]
mod tests {
    use super::*;
}
