pub type MultiKeyManagerResult<O> = Result<O, KeyManagementError>;

#[derive(Debug, PartialEq)]
pub enum KeyManagementError {
    /// None of the slots could be verified
    NoMatchingKeyFound,

    /// This error is yielded if the supplied key material is too short/weak
    BaseMaterialTooShort,

    /// If you see this error, please update your software
    RekeyRequired,

    /// A filesystem `std::io::Error`
    GeneralIO(String),

    /// A serialisation error occured
    RmpEncode(String),
    /// A deserialisation error occured
    RmpDecode(String),

    /// A constant time error occured inside a `ring::*` function
    GeneralRingError,

    /// Error occured during paper key strengthening
    InvalidPassword(argon2::password_hash::Error),

    /// An error related to Argon2 processing
    InvalidArgon2(argon2::Error),
}

impl From<ring::error::Unspecified> for KeyManagementError {
    fn from(_e: ring::error::Unspecified) -> Self {
        KeyManagementError::GeneralRingError
    }
}
impl From<argon2::password_hash::Error> for KeyManagementError {
    fn from(e: argon2::password_hash::Error) -> Self {
        KeyManagementError::InvalidPassword(e)
    }
}
impl From<argon2::Error> for KeyManagementError {
    fn from(e: argon2::Error) -> Self {
        KeyManagementError::InvalidArgon2(e)
    }
}

impl From<std::io::Error> for KeyManagementError {
    fn from(e: std::io::Error) -> Self {
        KeyManagementError::GeneralIO(format!("{e:?}"))
    }
}
impl From<rmp_serde::encode::Error> for KeyManagementError {
    fn from(e: rmp_serde::encode::Error) -> Self {
        KeyManagementError::RmpEncode(format!("{e:?}"))
    }
}
impl From<rmp_serde::decode::Error> for KeyManagementError {
    fn from(e: rmp_serde::decode::Error) -> Self {
        KeyManagementError::RmpDecode(format!("{e:?}"))
    }
}
