use crate::{LiveKey, LiveNonce, MultiKeyManagerResult, SaltString};
use serde::{Deserialize, Serialize};

pub const MASTER_KEY_LENGTH: usize = 256 / 8;
pub const MASTER_NONCE_LENGTH: usize = 96 / 8;
pub const MASTER_NONCE_FIELD_LENGTH: usize = 128 / 8;
pub const CHECK_VALUE_LENGTH: usize = 128 / 8;
pub const KEYSTREAM_LENGTH: usize =
    MASTER_KEY_LENGTH + MASTER_NONCE_FIELD_LENGTH + CHECK_VALUE_LENGTH;
pub const SALT_FIELD_LENGTH: usize = 128 / 8;

/// A key slot that can be stored on disk
#[derive(Serialize, Deserialize)]
pub struct Slot {
    encrypted_master_key: [u8; MASTER_KEY_LENGTH],
    encrypted_master_nonce: [u8; MASTER_NONCE_FIELD_LENGTH],
    check_value: [u8; CHECK_VALUE_LENGTH],
    /// This is the actual raw value, not b64 encoded
    salt: [u8; SALT_FIELD_LENGTH],
}

impl Slot {
    pub fn new(
        master_key: &LiveKey,
        master_nonce: &LiveNonce,
        salt: &SaltString,
        key_stream: &[u8],
    ) -> MultiKeyManagerResult<Self> {
        assert!(
            KEYSTREAM_LENGTH <= key_stream.len(),
            "key_stream must be at least {KEYSTREAM_LENGTH} Bytes long, was {}",
            key_stream.len()
        );
        assert_eq!(
            SALT_FIELD_LENGTH,
            calc_raw_length(salt),
            "salt length mismatch"
        );
        let (master_candidate, key_stream) = key_stream.split_at(MASTER_KEY_LENGTH);
        let (nonce_cadidate, check_value) = key_stream.split_at(MASTER_NONCE_FIELD_LENGTH);

        let encrypted_master_key = master_key ^ master_candidate;
        let encrypted_master_nonce = master_nonce ^ nonce_cadidate;

        let mut salt_buffer = [0; SALT_FIELD_LENGTH];
        // get the actual raw value
        // <[u8; 16]>::try_from(salt.as_bytes()) sadly gives the ascii encoded data
        salt.b64_decode(&mut salt_buffer)?;

        Ok(Slot {
            encrypted_master_key,
            encrypted_master_nonce,
            check_value: check_value
                .try_into()
                .expect("check_value length is mismatch"),
            salt: salt_buffer,
        })
    }

    pub fn update_xored_nonce(&mut self, old: &LiveNonce, new: u128) {
        let temp = old ^ &self.encrypted_master_nonce;
        for (e, (t, n)) in self
            .encrypted_master_nonce
            .iter_mut()
            .zip(temp.iter().zip(new.to_le_bytes().iter()))
        {
            *e = t ^ n;
        }
    }

    pub fn salt_string(&self) -> MultiKeyManagerResult<SaltString> {
        // yes this is confusing, PR to upstream has been submitted see:
        // https://github.com/RustCrypto/traits/pull/1216
        Ok(SaltString::b64_encode(&self.salt)?)
    }

    pub fn check_value_matches(&self, key_stream: &[u8]) -> bool {
        let (_head, check_value) =
            key_stream.split_at(MASTER_KEY_LENGTH + MASTER_NONCE_FIELD_LENGTH);
        self.check_value == check_value
    }

    pub fn decrypt_master_values(&self, key_stream: &[u8]) -> (LiveKey, LiveNonce) {
        let (master_xor, key_stream) = key_stream.split_at(MASTER_KEY_LENGTH);
        let (nonce_xor, _check_value) = key_stream.split_at(MASTER_NONCE_FIELD_LENGTH);

        (
            LiveKey::from_xor(
                &self.encrypted_master_key,
                master_xor.try_into().expect("Slot::decrypt_master_values"),
            ),
            LiveNonce::from_xor(
                &self.encrypted_master_nonce,
                nonce_xor.try_into().expect("Slot::decrypt_master_values"),
            ),
        )
    }

    pub fn id(&self) -> KeyId {
        const LENGTH: usize = 5;
        let mut s = String::with_capacity(LENGTH * 2);

        for b in &self.check_value[..LENGTH] {
            s += &*format!("{:02x}", b);
        }

        KeyId(s)
    }

    #[cfg(test)]
    pub(crate) fn __tests_encrypted_master_nonce(&self) -> [u8; 16] {
        self.encrypted_master_nonce.clone()
    }
}

impl std::fmt::Debug for Slot {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Slot")
            .field("id", &format_args!("{}", self.id().0))
            .finish()
    }
}

fn calc_raw_length(salt: &SaltString) -> usize {
    let l = salt.len() as f64;
    let r = l / 6.0;
    r.ceil() as usize * 4
}

/// probably change that
#[derive(Debug, PartialEq, Clone)]
pub struct KeyId(String);

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn calc_raw_length_22_to_16() {
        let salt = SaltString::new("5zUdLeTk0y/un5GN8Knwvg").unwrap();
        assert_eq!(16, calc_raw_length(&salt));
    }

    #[test]
    fn master_candidate() {
        let master_key = LiveKey {
            secret_key: [0x77; 32],
        };
        let master_candidate = [0xeeu8; 32];
        let encrypted_master_key = &master_key ^ &master_candidate;
        assert_eq!([0x99; 32], encrypted_master_key);
    }

    #[test]
    fn nonce_candidate() {
        let master_nonce = LiveNonce {
            counter: 0x77777777777777777777777777777777,
        };
        let nonce_cadidate = [0xeeu8; 16];
        let encrypted_master_nonce = &master_nonce ^ &nonce_cadidate;
        assert_eq!([0x99; MASTER_NONCE_FIELD_LENGTH], encrypted_master_nonce);
    }
}
